@@ -165,15 +165,6 @@ test_expect_success 'absolute path rejects the empty string' '
 	test_must_fail test-tool path-utils absolute_path ""
 '
 
-test_expect_success MINGW '<drive-letter>:\\abc is an absolute path' '
-	for letter in : \" C Z 1 ä
-	do
-		path=$letter:\\abc &&
-		absolute="$(test-tool path-utils absolute_path "$path")" &&
-		test "$path" = "$absolute" || return 1
-	done
-'
-
 test_expect_success 'real path rejects the empty string' '
 	test_must_fail test-tool path-utils real_path ""
 '
@@ -315,6 +306,8 @@ test_git_path GIT_COMMON_DIR=bar hooks/me                 bar/hooks/me
 test_git_path GIT_COMMON_DIR=bar config                   bar/config
 test_git_path GIT_COMMON_DIR=bar packed-refs              bar/packed-refs
 test_git_path GIT_COMMON_DIR=bar shallow                  bar/shallow
+test_git_path GIT_COMMON_DIR=bar common                   bar/common
+test_git_path GIT_COMMON_DIR=bar common/file              bar/common/file
 
 # In the tests below, $(pwd) must be used because it is a native path on
 # Windows and avoids MSYS's path mangling (which simplifies "foo/../bar" and
@@ -430,9 +423,6 @@ test_expect_success 'match .gitmodules' '
 		~1000000 \
 		~9999999 \
 		\
-		.gitmodules:\$DATA \
-		"gitmod~4 . :\$DATA" \
-		\
 		--not \
 		".gitmodules x"  \
 		".gitmodules .x" \
@@ -457,25 +447,7 @@ test_expect_success 'match .gitmodules' '
 		\
 		GI7EB~1 \
 		GI7EB~01 \
-		GI7EB~1X \
-		\
-		.gitmodules,:\$DATA
-'
-
-test_expect_success MINGW 'is_valid_path() on Windows' '
-       test-tool path-utils is_valid_path \
-		win32 \
-		"win32 x" \
-		../hello.txt \
-		C:\\git \
-		\
-		--not \
-		"win32 "  \
-		"win32 /x "  \
-		"win32."  \
-		"win32 . ." \
-		.../hello.txt \
-		colon:test
+		GI7EB~1X
 '
 
 test_done
